__all__ = [
	'ExpandContractions',
	'ContractContractions',
	'Emojify',
	'AddEmoji',
	'AddPositiveEmoji',
	'AddNegativeEmoji',
	'AddNeutralEmoji',
	'Demojify', 
	'RemoveEmoji', 
	'RemovePositiveEmoji',
	'RemoveNegativeEmoji',
	'RemoveNeutralEmoji',
	'ChangeLocation',
	'ChangeName',
	'InsertPositivePhrase',
	'InsertNegativePhrase',
	'RandomInsertion',
	'AddSentimentLink',
	'ImportLinkText',
	'AddNegation',
	'RemoveNegation',
	'RandomCharDel',
	'RandomCharInsert', 
	'RandomCharSubst', 
	'RandomCharSwap', 
	'RandomSwapQwerty',
	'ChangeNumber',
	'ChangeSynonym', 
	'ChangeAntonym', 
	'ChangeHyponym', 
	'ChangeHypernym',
	'WordDeletion', 
	'HomoglyphSwap', 
	'RandomSwap'
]